// permtype-getattrlist.m -- use getattrlist() to discover type and permisisons
//			     for a file.

/* compile with 
cc -g -o permtype-getattrlist permtype-getattrlist.m
*/

#import <sys/stat.h>    // for stat() and struct stat
#import <stdlib.h>      // for EXIT_SUCCESS
#import <stdio.h>       // for printf
#import <errno.h>       // for errno
#import <grp.h>         // for group file access routines
#import <pwd.h>         // for passwd file access routines
#import <sys/attr.h>    // for attribute structures
#import <unistd.h>      // for getattrlist()
#import <string.h>	// for memset()


// cheesy little lookup table for mapping perm value to the 
// familiar character string
static const char *g_perms[]  = {
    "---", "--x", "-w-", "-wx", "r--", "r-x", "rw-", "rwx"
};

typedef struct StatType {
    unsigned long	mask;
    const char	       *type;
} StatType;

static StatType g_types[] = {
    { S_IFREG, "Regular File" },
    { S_IFDIR, "Directory" },
    { S_IFLNK, "Symbolic Link" },
    { S_IFCHR, "Character Special Device" },
    { S_IFBLK, "Block Special Device" },
    { S_IFIFO, "FIFO" },
    { S_IFSOCK, "Socket" },
};


// structure of the data being returned by getattrlist()

typedef struct PermTypeAttributes {
    unsigned long	length;
    attrreference_t	name;
    struct timespec	modTime;
    struct timespec	changeTime;
    struct timespec	accessTime;
    uid_t		ownerId;
    gid_t		groupId;
#if __LITTLE_ENDIAN__
    mode_t		accessMask;
    short 		padding;
#else
    short 		padding;
    mode_t		accessMask;
#endif
    off_t		fileLogicalSize;
} PermTypeAttributes;


void displayInfo (const char *filename)
{
    struct attrlist attrList;

    // must clear out the structure first, otherwise we'll get
    // undefined results
    memset (&attrList, 0, sizeof(attrList));
    attrList.bitmapcount = ATTR_BIT_MAP_COUNT;

    // get the name, the permissions, and the stat-style times
    attrList.commonattr = ATTR_CMN_NAME
        | ATTR_CMN_MODTIME | ATTR_CMN_CHGTIME | ATTR_CMN_ACCTIME
        | ATTR_CMN_OWNERID | ATTR_CMN_GRPID | ATTR_CMN_ACCESSMASK;

    // also get the size
    attrList.fileattr = ATTR_FILE_TOTALSIZE;

    // the returned data cannot be larger than this, so we're safe
    // as far as silent data truncation goes
    char attrBuffer[sizeof(PermTypeAttributes) + NAME_MAX + 1];

    int result = getattrlist (filename, &attrList, attrBuffer, 
                              sizeof(attrBuffer), 0);

    if (result == -1) {
        fprintf (stderr, "error with getattrlist(%s) : %d / %s\n",
                 filename, errno, strerror(errno));
        return;
    }

    // overlay our structure on top of the returned bytes
    PermTypeAttributes *permAttributes;
    permAttributes = (PermTypeAttributes *) attrBuffer;

    // get to the string data from the attribute
    char *filenameFromAttrs;
    filenameFromAttrs = ((char *)&permAttributes->name) 
        + permAttributes->name.attr_dataoffset;

    printf ("%s:\n", filenameFromAttrs);

    printf ("  permissions: %s%s%s\n",
	    g_perms[(permAttributes->accessMask & S_IRWXU) >> 6],
	    g_perms[(permAttributes->accessMask & S_IRWXG) >> 3],
	    g_perms[(permAttributes->accessMask & S_IRWXO)]);

    // figure out the type
    StatType *scan, *stop;
    scan = g_types;
    stop = scan + (sizeof(g_types) / sizeof(StatType));

    while (scan < stop) {
	if ((permAttributes->accessMask & S_IFMT) == scan->mask) {
	    printf ("  type: %s\n", scan->type);
	    break;
	}
	scan++;
    }
    
    // any special bits sets?
    if ((permAttributes->accessMask & S_ISUID) == S_ISUID) {
	printf ("  set-uid!\n");
    }   
    if ((permAttributes->accessMask & S_ISGID) == S_ISUID) {
	printf ("  set-group-id!\n");
    }

    // file size
    printf ("  file is %lld bytes (%lld K)\n", 
	    permAttributes->fileLogicalSize,
	    permAttributes->fileLogicalSize / 1024);

    // owning user / group
    {
	struct passwd *passwd;
	struct group *group;

	passwd = getpwuid (permAttributes->ownerId);
	group = getgrgid (permAttributes->groupId);
	
	printf ("  user: %s (%d)\n", passwd->pw_name, permAttributes->ownerId);
	printf ("  group: %s (%d)\n", group->gr_name, permAttributes->groupId);
    }

    // now the dates
    {
	char buffer[1024];
	struct tm *tm;

	tm = localtime (&permAttributes->accessTime.tv_sec);
	strftime (buffer, 1024, "%c", tm);
	printf ("  last access: %s\n", buffer);

	tm = localtime (&permAttributes->modTime.tv_sec);
	strftime (buffer, 1024, "%c", tm);
	printf ("  last modification: %s\n", buffer);

	tm = localtime (&permAttributes->changeTime.tv_sec);
	strftime (buffer, 1024, "%c", tm);
	printf ("  last inode change: %s\n", buffer);
    }

    // double-space output
    printf ("\n");

} // displayInfo


int main (int argc, char *argv[])
{
    int i;

    if (argc == 1) {
	fprintf (stderr, "usage:  %s /path/to/file ... \n", argv[0]);
	exit (EXIT_FAILURE);
    }

    for (i = 1; i < argc; i++) {
	displayInfo (argv[i]);
    }

    exit (EXIT_SUCCESS);

} // main
